from sqlalchemy.sql import expression

from .. import db


class Users(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    email = db.Column(db.String(255), unique=True, nullable=False)
    enabled = db.Column(db.Boolean, server_default=expression.true())
    auth_providers = db.relationship('UserAuthProviders', backref='user')


import logging
from functools import wraps
from flask import request, jsonify
from .utils import decode_access_token
logger = logging.getLogger(__name__)

def token_required(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        logger.info("Accessing protected route with token_required decorator")
        auth_header = request.headers.get('Authorization')
        if not auth_header:
            logger.warning("Authorization header is missing")
            return jsonify({'message': 'Missing authorization token'}), 401

        try:
            token_type, token = auth_header.split()
            if token_type.lower() != 'bearer':
                logger.warning(f"Unexpected token type '{token_type}' in Authorization header")
                return jsonify({'message': 'Invalid token type. Expected Bearer'}), 401
        except ValueError:
            logger.error("Malformed Authorization header format")
            return jsonify({'message': 'Invalid Authorization header format. Expected: Bearer <token>'}), 401
        logger.debug(f"Authorization token received: {token}")
        payload = decode_access_token(token)
        if not payload:
            return jsonify({'message': 'Invalid token'}), 401
        payload.pop('exp', None)
        logger.debug(f"Token payload after decoding: {payload}")
        return f(*args, **payload)
    return decorated



import os

from dotenv import load_dotenv

dotenv_path = '../local.env'
load_dotenv(dotenv_path)

import flask_auth_wrapper
template_path = os.path.join(os.path.dirname(__file__), 'templates')
app = flask_auth_wrapper.create_app('flask_auth_wrapper.config.Config', template_folder=template_path)

if __name__ == '__main__':
    app.run(host='localhost', port=8080)



from .. import db


class UserAuthProviders(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id'))
    auth_provider_id = db.Column(db.Integer, db.ForeignKey('auth_providers.id'))
    name = db.Column(db.String(255))
    photo = db.Column(db.String)
    password = db.Column(db.String(255))
    tokens = db.relationship('Tokens', backref='user_auth_providers')




import time
import uuid

import jwt
from flask import current_app


def create_access_token(user, user_auth_provider):
    payload = {'ua_id': user_auth_provider.id,
               'user_email': user.email,
               'user_name': user_auth_provider.name,
               'exp': time.time() + 3600
               }  # 1 hour expiry
    token = jwt.encode(payload, current_app.config['SECRET_KEY'], algorithm='HS256')
    return token


def generate_refresh_token():
    return uuid.uuid4().hex


def decode_access_token(token):
    try:
        payload = jwt.decode(token, current_app.config['SECRET_KEY'], algorithms=['HS256'])
        return payload
    except jwt.ExpiredSignatureError:
        return None
    except jwt.InvalidTokenError:
        return None



import logging

from authlib.integrations.base_client import MismatchingStateError
from flask import Blueprint, request, jsonify, session, redirect, url_for, render_template



from . import oauth, db
from .decorators import token_required
from .exceptions import UserNotFoundException, InvalidProviderError, InvalidRefreshTokenError, ValidationError
from .models.tokens_model import Tokens
from .services.services import find_user_and_provider, update_user_auth_provider, add_token, validate_refresh_token, \
    revoke_tokens, get_user_auth_provider_by_token, get_user_provider_by_user_auth_provider, get_user_details
from .utils import create_access_token, generate_refresh_token

logger = logging.getLogger(__name__)
auth_bp = Blueprint('auth', __name__, url_prefix='/auth')


@auth_bp.errorhandler(Exception)
def handle_auth_error(ex):
    logger.error(f"Exception in endpoint {request.endpoint}: {ex}", exc_info=True)

    def response(code: int, message: str, details: dict = None, error_type: str = None):
        response_data = {
            'code': code,
            'message': message
        }
        if details:
            response_data['details'] = details
        if error_type:
            response_data['error_type'] = error_type
        return jsonify(response_data), code

    if isinstance(ex, MismatchingStateError):
        return response(code=403, message=ex.description, error_type='CSRFError')


    if isinstance(ex, (InvalidProviderError, InvalidRefreshTokenError, ValidationError, UserNotFoundException)):
        return response(code=ex.code, message=ex.message, details=ex.details, error_type=ex.__class__.__name__)

    # Handle errors specific to the auth blueprint
    return response(code=500, message='something went wrong...')


@auth_bp.route('/')
def homepage():
    user = session.get('user')
    logger.info(f"Homepage accessed. User: {user}")
    return render_template('home.html', user=user)


@auth_bp.route('/login/<provider>')
def login(provider):
    logger.info(f"Login attempt for provider: {provider}")
    client = oauth.create_client(provider)
    if not client:
        logger.error(f"OAuth client for provider '{provider}' not found")
        raise InvalidProviderError(f"OAuth client for provider '{provider}' not found.")
    redirect_uri = url_for('auth.auth', provider=provider, _external=True)
    logger.debug(f"Redirect URI generated: {redirect_uri}")
    return client.authorize_redirect(redirect_uri)


@auth_bp.route('/login', methods=['POST'])
def local_login():
    email = request.form.get('email')
    password = request.form.get('password')
    logger.info(f"Local login attempt. Email: {email}")

    if not email or not password:
        logger.warning("Missing email or password")
        return jsonify({'message': 'Missing email or password'}), 400

    existing_user, auth_provider = find_user_and_provider(email, 'local')
    user_auth_provider = update_user_auth_provider(user=existing_user, provider=auth_provider       )

    if not existing_user or password != user_auth_provider.password:
        logger.error("Invalid username or password")
        raise UserNotFoundException('Invalid username or password')

    access_token = create_access_token(existing_user, user_auth_provider)
    refresh_token = generate_refresh_token()
    logger.info(f"Generated tokens for user: {email}")

    revoked_tokens = revoke_tokens(user_auth_provider=user_auth_provider)

    token = add_token(
        user_auth_provider=user_auth_provider,
        access_token=access_token,
        refresh_token=refresh_token
    )

    db.session.add(user_auth_provider)
    db.session.add(token)
    db.session.commit()

    return jsonify({'access_token': access_token, 'refresh_token': refresh_token})


@auth_bp.route('<provider>')
def auth(provider):
    logger.info(f"OAuth callback for provider: {provider}")
    client = oauth.create_client(provider)
    token = client.authorize_access_token()
    user_info = token.get('userinfo') or client.userinfo()

    user_email = user_info.get('email')

    if not user_email:
        return redirect('/')

    logger.info(f"User authenticated with email: {user_email}")

    # find user and provider
    existing_user, auth_provider = find_user_and_provider(user_email, provider)

    user_auth_provider = update_user_auth_provider(user=existing_user, provider=auth_provider, user_info=user_info)

    access_token = create_access_token(existing_user, user_auth_provider)
    refresh_token = generate_refresh_token()

    logger.info(f"Generated tokens for user: {user_email}")

    revoked_tokens = revoke_tokens(user_auth_provider=user_auth_provider)

    token = add_token(
        user_auth_provider=user_auth_provider,
        access_token=access_token,
        refresh_token=refresh_token
    )

    db.session.add(user_auth_provider)
    db.session.add(token)
    db.session.commit()

    return jsonify({'access_token': access_token, 'refresh_token': refresh_token})


@auth_bp.route('/refresh', methods=['POST'])
def refresh():
    auth_header = request.headers.get('Authorization', None)
    if not auth_header:
        logger.warning("Missing Authorization header")
        return jsonify({'message': 'Authorization header missing'}), 400

    try:
        refresh_token = auth_header.split()[1]
    except IndexError:
        logger.error("Invalid Authorization header format")
        return jsonify({'message': 'Invalid Authorization header format. Expected: Bearer <token>'}), 400

    refresh_token = request.headers.get('Authorization').split()[1]
    if not refresh_token:
        return jsonify({'message': 'Missing refresh token'}), 400

    logger.info("Refreshing tokens")
    _token = validate_refresh_token(refresh_token)
    user_auth_provider = get_user_auth_provider_by_token(_token)
    user, provider = get_user_provider_by_user_auth_provider(user_auth_provider)
    if not user:
        logger.error("User not found or disabled")
        raise UserNotFoundException('User not found or disabled')

    new_access_token = create_access_token(user, user_auth_provider)
    new_refresh_token = generate_refresh_token()

    logger.info(f"Generated new tokens for user: {user.email}")

    revoke_tokens(token=_token)
    new_token = add_token(
        user_auth_provider=user_auth_provider,
        access_token=new_access_token,
        refresh_token=new_refresh_token
    )
    db.session.add(new_token)

    db.session.commit()

    return jsonify({'access_token': new_access_token, 'refresh_token': new_refresh_token})


@auth_bp.route('/whoami')
@token_required
def whoami(ua_id, user_email, **kwargs):
    logger.info(f"Received request for who_am_i")
    user_details = get_user_details(ua_id, user_email)
    _user_name = user_details.get('name')

    response = dict(
        message=f'Welcome {_user_name}!',
        details=user_details
    )
    return jsonify(response)


@auth_bp.route('/logout')
@token_required
def logout(user, **kwargs):
    auth_header = request.headers.get('Authorization')
    request_token = auth_header.split()[1]
    _token = Tokens.query.filter_by(token=request_token)
    revoke_tokens(token=_token)
    db.session.commit()
    return redirect('/')



class AppException(Exception):
    default_code = 500
    default_message = "Something went wrong..."
    default_details = {}

    def __init__(self, message: str = None,  code: int = None, details: dict = None) -> None:
        self.code = code or self.default_code
        self.message = message or self.default_message
        self.details = details or self.default_details
        super().__init__(message)


class UserNotFoundException(AppException):
    default_code = 404
    default_message = 'User Not Found!'
    pass


class AuthProviderNotFoundException(AppException):
    pass


class ValidationError(AppException):
    default_code = 400
    default_message = "Validation Error!"


class InvalidRefreshTokenError(ValidationError):
    pass


class InvalidProviderError(ValidationError):
    pass



from flask import Flask
from flask_sqlalchemy import SQLAlchemy
from authlib.integrations.flask_client import OAuth

db = SQLAlchemy()
oauth = OAuth()


def register_oauth_clients(app):
    for provider, config in app.config['OAUTH2_PROVIDERS'].items():
        oauth.register(
            name=provider,
            client_id=config.get('client_id'),
            client_secret=config.get('client_secret'),
            server_metadata_url=config.get('server_metadata_url'),
            api_base_url=config.get('api_base_url'),
            request_token_url=config.get('request_token_url'),
            access_token_url=config.get('access_token_url'),
            authorize_url=config.get('authorize_url'),
            userinfo_endpoint=config.get('userinfo_endpoint'),
            userinfo_compliance_fix=config.get('userinfo_compliance_fix'),
            client_kwargs=config.get('client_kwargs', {}),
            jwks_uri=config.get('jwks_uri')
        )


def create_app(config_object, **kwargs):
    app = Flask(__name__, **kwargs)
    app.config.from_object(config_object)

    db.init_app(app)
    oauth.init_app(app)

    register_oauth_clients(app)

    with app.app_context():
        from .auth import auth_bp
        app.register_blueprint(auth_bp)

        db.create_all()

    return app



import os


class Config:
    SECRET_KEY = os.getenv('SECRET_KEY')
    SQLALCHEMY_DATABASE_URI = os.getenv('SQLALCHEMY_DATABASE_URI')
    SQLALCHEMY_TRACK_MODIFICATIONS = False

    OAUTH2_PROVIDERS = {
        'google': {
            'client_id': os.getenv('GOOGLE_CLIENT_ID'),
            'client_secret': os.getenv('GOOGLE_CLIENT_SECRET'),
            'server_metadata_url': os.getenv('GOOGLE_METADATA_URL'),
            'client_kwargs': {'scope': 'openid email profile'}
        },
        'twitter': {
            'api_base_url': os.getenv('OAUTH2_TWITTER_API_BASE_URL'),
            'request_token_url': os.getenv('OAUTH2_TWITTER_REQUEST_TOKEN_URL'),
            'access_token_url': os.getenv('OAUTH2_TWITTER_ACCESS_TOKEN_URL'),
            'authorize_url': os.getenv('OAUTH2_TWITTER_AUTHORIZE_URL'),
            'userinfo_endpoint': os.getenv('OAUTH2_TWITTER_USERINFO_ENDPOINT'),
            'userinfo_compliance_fix': os.getenv('OAUTH2_TWITTER_USERINFO_COMPLIANCE_FIX')
        },
        'azure': {
            'client_id': os.getenv('AZURE_CLIENT_ID'),
            'client_secret': os.getenv('AZURE_CLIENT_SECRET'),
            'authorize_url': os.getenv('AZURE_AUTH_URL'),
            'access_token_url': os.getenv('AZURE_TOKEN_URL'),
            'client_kwargs': {'scope': 'openid email profile'},
            'jwks_uri': os.getenv('AZURE_JWKS_URL'),

        }
    }

    @staticmethod
    def init_app(app):
        pass  # Add any additional initialization here if needed



from datetime import datetime

from .. import db


class Tokens(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    user_auth_provider_id = db.Column(db.Integer, db.ForeignKey('user_auth_providers.id'))
    token = db.Column(db.Text, unique=True, nullable=False)
    refresh_token = db.Column(db.Text, unique=True, nullable=False)
    revoked = db.Column(db.Boolean, default=False)
    created_at = db.Column(db.DateTime(timezone=True), default=datetime.utcnow)



from setuptools import setup, find_packages

# Define package metadata
NAME = "flask-oauth-guard"
VERSION = "0.0.8"
AUTHOR = "Raghav Sethi"
AUTHOR_EMAIL = "work.raghavsethi@gmail.com"
DESCRIPTION = "A Flask extension to add Authentication and Authorization, supports OAuth2"
URL = "https://github.com/05rs/flask-auth"
LICENSE = "Apache"

# Define dependencies (if any)
INSTALL_REQUIRES = [
    "Flask",
    "authlib",
    "requests",
    "urllib3==1.26.15",
    "flask-sqlalchemy",
    "flask-wtf",
    "psycopg2-binary",
    "PyJWT",
    "python-dotenv==1.0.1"
]



# Find package structure
PACKAGES = ["flask_auth_wrapper", "flask_auth_wrapper.models", "flask_auth_wrapper.services"]

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


# Create setup arguments
setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    description=DESCRIPTION,
    url=URL,
    license=LICENSE,
    install_requires=INSTALL_REQUIRES,
    packages=PACKAGES,
    long_description=long_description,
    long_description_content_type="text/markdown",
)

# python3 setup.py sdist bdist_wheel



from datetime import datetime, timezone

from ..exceptions import InvalidRefreshTokenError, UserNotFoundException, \
    AuthProviderNotFoundException
from ..models.auth_providers_model import AuthProviders
from ..models.tokens_model import Tokens
from ..models.user_auth_providers_model import UserAuthProviders
from ..models.users_model import Users


def validate_refresh_token(refresh_token):
    _token = Tokens.query.filter_by(refresh_token=refresh_token, revoked=False).first()
    if not _token:
        raise InvalidRefreshTokenError(message='Invalid Refresh Token!')
    return _token


def find_user_and_provider(user_email, provider):
    existing_user = Users.query.filter_by(email=user_email, enabled=True).first()
    if not existing_user:
        raise UserNotFoundException(message=f"User '{user_email}' not found or disabled!")

    auth_provider = AuthProviders.query.filter_by(name=provider).first()
    if not auth_provider:
        raise AuthProviderNotFoundException(message=f"Auth provider '{provider}' not supported")

    return existing_user, auth_provider


def get_user_auth_provider_by_token(token: Tokens):
    return UserAuthProviders.query.get(token.user_auth_provider_id)


def get_user_provider_by_user_auth_provider(user_auth_provider: UserAuthProviders):
    user = Users.query.filter_by(id=user_auth_provider.user_id, enabled=True).first()
    provider = AuthProviders.query.get(user_auth_provider.auth_provider_id)
    return user, provider


def get_profile_avatar(name):
    if name is None:
        return ""
    return f'https://ui-avatars.com/api/?background=random&color=%23ffffff&size=95&length=2&name={name}&font-size=0.55'

def update_user_auth_provider(user, provider, user_info={}):
    user_auth_provider = UserAuthProviders.query.filter_by(user_id=user.id,auth_provider_id=provider.id).first()

    if not user_auth_provider:
        user_auth_provider = UserAuthProviders(
            user_id=user.id,
            auth_provider_id=provider.id,
            name=user_info.get('name'),
            photo=user_info.get('picture', get_profile_avatar(user_info.get('name', None)))
        )
    else:
        if user_info.get('name'):
            user_auth_provider.name = user_info.get('name')

        # Update the photo if a picture is provided or if there's no existing photo
        if user_info.get('picture') or (user_auth_provider.name and not user_auth_provider.photo):
            user_auth_provider.photo = user_info.get('picture', None) or get_profile_avatar(user_auth_provider.name)

    return user_auth_provider


def revoke_tokens(token: Tokens = None, user_auth_provider: UserAuthProviders = None):
    if token:
        token.revoked = True
        return token
    if user_auth_provider:
        tokens = Tokens.query.filter_by(user_auth_provider_id=user_auth_provider.id, revoked=False)
        tokens.update({'revoked': True}, synchronize_session=False)
        return tokens


def add_token(user_auth_provider, access_token, refresh_token):
    return Tokens(
        user_auth_providers=user_auth_provider,
        token=access_token,
        refresh_token=refresh_token,
        revoked=False,
        created_at=datetime.now(tz=timezone.utc),
    )


def get_user_details(ua_id, user_email):
    user_auth_provider = UserAuthProviders.query.filter_by(id=ua_id).first()
    return dict(
        name=user_auth_provider.name,
        email=user_email,
        photo=user_auth_provider.photo
    )

